@@ -230,6 +230,45 @@ fn query_results_with_externals(query: Query) -> (QueryResults, MockExternal) {
     )
 }
 
+/// equality test for polar expressions that takes symmetric operators
+/// into account, eg. a = b == b = a
+fn commute_ops(u: &Value, v: &Value) -> bool {
+    fn a2p(a: &[Term]) -> (&Value, &Value) {
+        (a[0].value(), a[1].value())
+    }
+    match (u.as_expression(), v.as_expression()) {
+        (
+            Ok(Operation {
+                operator: op_a,
+                args: arg_a,
+            }),
+            Ok(Operation {
+                operator: op_b,
+                args: arg_b,
+            }),
+        ) if op_a == op_b && arg_a.len() == arg_b.len() => {
+            let op = *op_a;
+            if arg_a.len() == 2
+                && (op == Operator::Unify
+                    || op == Operator::Eq
+                    || op == Operator::Neq
+                    || op == Operator::And
+                    || op == Operator::Or)
+            {
+                let (a, b) = (a2p(arg_a), a2p(arg_b));
+                commute_ops(a.0, b.0) && commute_ops(a.1, b.1)
+                    || commute_ops(a.0, b.1) && commute_ops(a.1, b.0)
+            } else {
+                arg_a
+                    .iter()
+                    .enumerate()
+                    .all(|(i, x)| commute_ops(arg_b[i].value(), x.value()))
+            }
+        }
+        _ => u == v,
+    }
+}
+
 #[track_caller]
 #[must_use = "test results need to be asserted"]
 fn eval(p: &mut Polar, query_str: &str) -> bool {
@@ -570,9 +609,9 @@ fn test_constants() -> TestResult {
     let mut p = Polar::new();
     {
         let mut kb = p.kb.write().unwrap();
-        kb.constant(sym!("one"), term!(1));
-        kb.constant(sym!("two"), term!(2));
-        kb.constant(sym!("three"), term!(3));
+        kb.constant(sym!("one"), term!(1))?;
+        kb.constant(sym!("two"), term!(2))?;
+        kb.constant(sym!("three"), term!(3))?;
     }
     p.load_str(
         r#"one(x) if one = one and one = x and x < two;
@@ -602,6 +641,8 @@ fn test_not() -> TestResult {
     qnull(&mut p, "even(3)");
     qeval(&mut p, "not even(3)");
 
+    p.clear_rules();
+
     p.load_str(
         r#"f(x) if not a(x);
            a(1);
@@ -618,6 +659,8 @@ fn test_not() -> TestResult {
     qeval(&mut p, "g(x) and x=3");
     qeval(&mut p, "x=3 and g(x)");
 
+    p.clear_rules();
+
     p.load_str("h(x) if not (not (x = 1 or x = 3) or x = 3);")?;
     qeval(&mut p, "h(1)");
     qnull(&mut p, "h(2)");
@@ -625,10 +668,14 @@ fn test_not() -> TestResult {
 
     qeval(&mut p, "d = {x: 1} and not d.x = 2");
 
+    p.clear_rules();
+
     // Negate And with unbound variable.
     p.load_str("i(x,y) if not (y = 2 and x = 1);")?;
     qvar(&mut p, "i(2,y)", "y", values![sym!("y")]);
 
+    p.clear_rules();
+
     // Negate Or with unbound variable.
     p.load_str("j(x,y) if not (y = 2 or x = 1);")?;
     qeval(&mut p, "j(2, y)");
@@ -753,8 +800,12 @@ fn test_or() -> TestResult {
     qnull(&mut p, "f(2)");
     qeval(&mut p, "f(3)");
 
+    p.clear_rules();
+
     p.load_str(
         r#"g(x) if a(x) or b(x) or c(x);
+           a(1);
+           b(3);
            c(5);"#,
     )?;
     qvar(&mut p, "g(x)", "x", values![1, 3, 5]);
@@ -804,10 +855,14 @@ fn test_non_instance_specializers() -> TestResult {
     qeval(&mut p, "f(1)");
     qnull(&mut p, "f(2)");
 
+    p.clear_rules();
+
     p.load_str("g(x: 1, y: [x]) if y = [1];")?;
     qeval(&mut p, "g(1, [1])");
     qnull(&mut p, "g(1, [2])");
 
+    p.clear_rules();
+
     p.load_str("h(x: {y: y}, x.y) if y = 1;")?;
     qeval(&mut p, "h({y: 1}, 1)");
     qnull(&mut p, "h({y: 1}, 2)");
@@ -987,7 +1042,7 @@ fn test_make_external() -> TestResult {
 #[test]
 fn test_external_call() -> TestResult {
     let p = Polar::new();
-    p.register_constant(sym!("Foo"), term!(true));
+    p.register_constant(sym!("Foo"), term!(true))?;
     let mut foo_lookups = vec![term!(1)];
 
     let q = p.new_query("(new Foo()).bar(1, a: 2, b: 3) = 1", false)?;
@@ -1037,6 +1092,8 @@ fn test_comparisons() -> TestResult {
     qeval(&mut p, "lt(\"aa\",\"ab\")");
     qnull(&mut p, "lt(\"aa\",\"aa\")");
 
+    p.clear_rules();
+
     // <=
     p.load_str("leq(x, y) if x <= y;")?;
     qeval(&mut p, "leq(1,1)");
@@ -1050,6 +1107,8 @@ fn test_comparisons() -> TestResult {
     qeval(&mut p, "leq(\"aa\",\"ab\")");
     qnull(&mut p, "leq(\"ab\",\"aa\")");
 
+    p.clear_rules();
+
     // >
     p.load_str("gt(x, y) if x > y;")?;
     qnull(&mut p, "gt(1,1)");
@@ -1062,6 +1121,8 @@ fn test_comparisons() -> TestResult {
     qeval(&mut p, "gt(\"ab\",\"aa\")");
     qnull(&mut p, "gt(\"aa\",\"aa\")");
 
+    p.clear_rules();
+
     // >=
     p.load_str("geq(x, y) if x >= y;")?;
     qeval(&mut p, "geq(1,1)");
@@ -1075,6 +1136,8 @@ fn test_comparisons() -> TestResult {
     qeval(&mut p, "geq(\"ab\",\"aa\")");
     qeval(&mut p, "geq(\"aa\",\"aa\")");
 
+    p.clear_rules();
+
     // ==
     p.load_str("eq(x, y) if x == y;")?;
     qeval(&mut p, "eq(1,1)");
@@ -1091,6 +1154,9 @@ fn test_comparisons() -> TestResult {
     qnull(&mut p, "eq(9007199254740992,9007199254740994.0)"); // distinguishable
     qeval(&mut p, "eq(\"aa\", \"aa\")");
     qnull(&mut p, "eq(\"ab\", \"aa\")");
+    qeval(&mut p, "eq(bob, bob)");
+
+    p.clear_rules();
 
     // !=
     p.load_str("neq(x, y) if x != y;")?;
@@ -1103,8 +1169,6 @@ fn test_comparisons() -> TestResult {
     qnull(&mut p, "neq(\"aa\", \"aa\")");
     qeval(&mut p, "neq(\"ab\", \"aa\")");
 
-    qeval(&mut p, "eq(bob, bob)");
-
     qeval(&mut p, "1.0 == 1");
     qeval(&mut p, "0.99 < 1");
     qeval(&mut p, "1.0 <= 1");
@@ -1512,6 +1576,48 @@ fn test_unknown_specializer_warning() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_and_or_warning() -> TestResult {
+    let p = Polar::new();
+
+    // free-standing OR is fine
+    p.load_str("f(x) if x > 1 or x < 3;")?;
+
+    p.clear_rules();
+
+    // OR with explicit parenthesis is fine (old behaviour)
+    p.load_str("f(x) if x = 1 and (x > 1 or x < 3);")?;
+
+    p.clear_rules();
+
+    // OR with parenthesized AND is fine (new default)
+    p.load_str("f(x) if (x = 1 and x > 1) or x < 3;")?;
+
+    p.clear_rules();
+
+    // Add whitespace to make sure it can find parentheses wherever they are
+    p.load_str("f(x) if (\n\t    x = 1 and  x > 1) or x < 3;")?;
+
+    p.clear_rules();
+    p.load_str("f(x) if x = 1 and x > 1 or x < 3;")?;
+    let msg = p.next_message().unwrap();
+    assert!(matches!(&msg.kind, MessageKind::Warning));
+    assert_eq!(
+        &msg.msg,
+        "Expression without parentheses could be ambiguous. \nPrior to 0.20, `x and y or z` would parse as `x and (y or z)`. \nAs of 0.20, it parses as `(x and y) or z`, matching other languages. \n\n\n001: f(x) if x = 1 and x > 1 or x < 3;\n             ^"
+    );
+
+    p.clear_rules();
+    p.load_str("f(x) if x = 1 or x > 1 and x < 3;")?;
+    let msg = p.next_message().unwrap();
+    assert!(matches!(&msg.kind, MessageKind::Warning));
+    assert_eq!(
+        &msg.msg,
+        "Expression without parentheses could be ambiguous. \nPrior to 0.20, `x and y or z` would parse as `x and (y or z)`. \nAs of 0.20, it parses as `(x and y) or z`, matching other languages. \n\n\n001: f(x) if x = 1 or x > 1 and x < 3;\n                      ^"
+    );
+    Ok(())
+}
+
 #[test]
 fn test_print() -> TestResult {
     // TODO: If POLAR_LOG is on this test will fail.
@@ -1572,6 +1678,8 @@ fn test_rest_vars() -> TestResult {
     qeval(&mut p, "not member(4, [1,2,3])");
     qvar(&mut p, "member(x, [1,2,3])", "x", values![1, 2, 3]);
 
+    p.clear_rules();
+
     p.load_str(
         r#"append([], x, x);
            append([first, *rest], x, [first, *tail]) if append(rest, x, tail);"#,
@@ -1583,6 +1691,11 @@ fn test_rest_vars() -> TestResult {
     qeval(&mut p, "append([1,2,3], [], [1,2,3])");
     qeval(&mut p, "not append([1,2,3], [4], [1,2,3])");
 
+    qeval(
+        &mut p,
+        "a = [1, *b] and b = [2, *c] and c=[3] and 1 in a and 2 in a and 3 in a",
+    );
+
     let a = &var(&mut p, "[*_] in [*a] and [*b] in [*_] and b = 1", "a")[0];
     // check that a isn't bound to [b]
     assert!(!matches!(a, Value::List(b) if matches!(b[0].value(), Value::Number(_))));
@@ -1592,8 +1705,8 @@ fn test_rest_vars() -> TestResult {
 #[test]
 fn test_circular_data() -> TestResult {
     let mut p = Polar::new();
-    qeval(&mut p, "x = [x]");
-    qeval(&mut p, "y = {y:y}");
+    qeval(&mut p, "x = [x] and x in x");
+    qeval(&mut p, "y = {y:y} and [\"y\", y] in y");
     qruntime!(
         "x = [x, y] and y = [y, x] and x = y",
         RuntimeError::StackOverflow { .. }
@@ -1601,6 +1714,46 @@ fn test_circular_data() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_data_filtering_dict_specializers() -> TestResult {
+    let pol_a = "allow(x, \"read\", _y: { x: x });";
+    let pol_b = "allow(x, \"read\", _y) if x = _y.x;";
+    let query = "allow(\"gwen\", \"read\", x)";
+    let p = Polar::new();
+    p.load_str(pol_a)?;
+    let mut res_a = query_results!(p.new_query(query, false)?);
+    p.clear_rules();
+    p.load_str(pol_b)?;
+    let mut res_b = query_results!(p.new_query(query, false)?);
+    assert_eq!(res_a.len(), 1);
+    assert_eq!(res_b.len(), 1);
+    let key = sym!("x");
+    let res_a = res_a[0].0.remove(&key).unwrap();
+    let res_b = res_b[0].0.remove(&key).unwrap();
+    assert!(commute_ops(&res_a, &res_b));
+    Ok(())
+}
+
+#[test]
+fn test_data_filtering_pattern_specializers() -> TestResult {
+    let pol_a = "allow(x, \"read\", _y: Dictionary{ x: x });";
+    let pol_b = "allow(x, \"read\", _y: Dictionary) if x = _y.x;";
+    let query = "allow(\"gwen\", \"read\", x)";
+    let p = Polar::new();
+    p.load_str(pol_a)?;
+    let mut res_a = query_results!(p.new_query(query, false)?);
+    p.clear_rules();
+    p.load_str(pol_b)?;
+    let mut res_b = query_results!(p.new_query(query, false)?);
+    assert_eq!(res_a.len(), 1);
+    assert_eq!(res_b.len(), 1);
+    let key = sym!("x");
+    let res_a = res_a[0].0.remove(&key).unwrap();
+    let res_b = res_b[0].0.remove(&key).unwrap();
+    assert!(commute_ops(&res_a, &res_b));
+    Ok(())
+}
+
 #[test]
 fn test_in_op() -> TestResult {
     let mut p = Polar::new();
@@ -1657,6 +1810,15 @@ fn test_in_op() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_head_patterns() -> TestResult {
+    let p = Polar::new();
+    p.load_str("f(x: Integer, _: Dictionary{x:x});")?;
+    let results = query_results!(p.new_query("f(x, {x:9})", false)?);
+    assert_eq!(results[0].0[&sym!("x")], value!(9));
+    Ok(())
+}
+
 #[test]
 fn test_matches() {
     let mut p = Polar::new();
@@ -1671,11 +1833,30 @@ fn test_matches() {
 }
 
 #[test]
-fn test_keyword_bug() {
-    qparse!("g(a) if a.new(b);", ParseError::ReservedWord { .. });
-    qparse!("f(a) if a.in(b);", ParseError::ReservedWord { .. });
+fn test_keyword_call() {
     qparse!("cut(a) if a;", ParseError::ReservedWord { .. });
     qparse!("debug(a) if a;", ParseError::ReservedWord { .. });
+    qparse!(
+        "foo(debug) if debug = 1;",
+        ParseError::UnrecognizedToken { .. }
+    );
+}
+
+#[test]
+fn test_keyword_dot() -> TestResult {
+    // field accesses of reserved words are allowed
+    let mut p = Polar::new();
+    p.load_str(
+        r#"
+        f(a, b) if a.in(b);
+        g(a, b) if a.new(b);
+    "#,
+    )?;
+    qeval(
+        &mut p,
+        "x = {debug: 1, new: 2, type: 3} and x.debug + x.new = x.type",
+    );
+    Ok(())
 }
 
 /// Test that rule heads work correctly when unification or specializers are used.
@@ -1693,7 +1874,7 @@ fn test_unify_rule_head() -> TestResult {
     );
 
     let p = Polar::new();
-    p.register_constant(sym!("Foo"), term!(true));
+    p.register_constant(sym!("Foo"), term!(true))?;
     p.load_str(
         r#"f(_: Foo{a: 1}, x) if x = 1;
            g(_: Foo{a: Foo{a: 1}}, x) if x = 1;"#,
@@ -1716,7 +1897,7 @@ fn test_cut() -> TestResult {
     p.load_str(
         r#"a(x) if x = 1 or x = 2;
            b(x) if x = 3 or x = 4;
-           bcut(x) if x = 3 or x = 4 and cut;
+           bcut(x) if (x = 3 or x = 4) and cut;
            c(a, b) if a(a) and b(b) and cut;
            c_no_cut(a, b) if a(a) and b(b);
            c_partial_cut(a, b) if a(a) and bcut(b);
@@ -1751,6 +1932,8 @@ fn test_cut() -> TestResult {
         values![[1, 3], [1, 4]],
     );
 
+    p.clear_rules();
+
     p.load_str("f(x) if (x = 1 and cut) or x = 2;")?;
     qvar(&mut p, "f(x)", "x", values![1]);
     qeval(&mut p, "f(1)");
@@ -1769,6 +1952,8 @@ fn test_forall() -> TestResult {
     qeval(&mut p, "all_ones([1, 1, 1])");
     qnull(&mut p, "all_ones([1, 2, 1])");
 
+    p.clear_rules();
+
     p.load_str("not_ones(l) if forall(item in l, item != 1);")?;
     qnull(&mut p, "not_ones([1])");
     qeval(&mut p, "not_ones([2, 3, 4])");
@@ -1779,6 +1964,8 @@ fn test_forall() -> TestResult {
     qeval(&mut p, "forall(x = 1, x = 1)");
     qeval(&mut p, "forall(x in [2, 3, 4], x > 1)");
 
+    p.clear_rules();
+
     p.load_str(
         r#"g(1);
            g(2);
@@ -1786,14 +1973,16 @@ fn test_forall() -> TestResult {
     )?;
     qeval(&mut p, "forall(g(x), x in [1, 2, 3])");
 
+    p.clear_rules();
+
     p.load_str(
-        r#"allow(_: {x: 1}, y) if y = 1;
-           allow(_: {y: 1}, y) if y = 2;
-           allow(_: {z: 1}, y) if y = 3;"#,
+        r#"test(_: {x: 1}, y) if y = 1;
+           test(_: {y: 1}, y) if y = 2;
+           test(_: {z: 1}, y) if y = 3;"#,
     )?;
     qeval(
         &mut p,
-        "forall(allow({x: 1, y: 1, z: 1}, y), y in [1, 2, 3])",
+        "forall(test({x: 1, y: 1, z: 1}, y), y in [1, 2, 3])",
     );
     Ok(())
 }
@@ -1914,9 +2103,9 @@ fn test_numeric_applicability() -> TestResult {
     let nan1 = f64::NAN;
     let nan2 = f64::from_bits(f64::NAN.to_bits() | 1);
     assert!(eps.is_normal() && nan1.is_nan() && nan2.is_nan());
-    p.register_constant(sym!("eps"), term!(eps));
-    p.register_constant(sym!("nan1"), term!(nan1));
-    p.register_constant(sym!("nan2"), term!(nan2));
+    p.register_constant(sym!("eps"), term!(eps))?;
+    p.register_constant(sym!("nan1"), term!(nan1))?;
+    p.register_constant(sym!("nan2"), term!(nan2))?;
     p.load_str(
         r#"f(0);
            f(1);
@@ -1942,7 +2131,7 @@ fn test_numeric_applicability() -> TestResult {
     qeval(&mut p, "f(9223372036854775807)");
     qeval(&mut p, "f(-9223372036854775807)");
     qeval(&mut p, "f(9223372036854776000.0)");
-    qnull(&mut p, "f(nan1)");
+    qeval(&mut p, "f(nan1)");
     qnull(&mut p, "f(nan2)");
     Ok(())
 }
@@ -2125,7 +2314,7 @@ fn test_builtin_iterables() {
 /// despite argument not matching lookup result
 fn test_lookup_in_rule_head() -> TestResult {
     let p = Polar::new();
-    p.register_constant(sym!("Foo"), term!(true));
+    p.register_constant(sym!("Foo"), term!(true))?;
     p.load_str(r#"test(foo: Foo, foo.bar());"#)?;
 
     let good_q = p.new_query("test(new Foo(), 1)", false)?;
@@ -2140,3 +2329,55 @@ fn test_lookup_in_rule_head() -> TestResult {
     assert_eq!(results.len(), 0);
     Ok(())
 }
+
+#[test]
+fn test_default_rule_types() -> TestResult {
+    let p = Polar::new();
+    // This should fail
+    let e = p
+        .load_str(r#"has_permission("leina", "eat", "food");"#)
+        .expect_err("Expected validation error");
+    assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    let e = p
+        .load_str(r#"has_role("leina", "eater", "food");"#)
+        .expect_err("Expected validation error");
+    assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    let e = p
+        .load_str(r#"has_relation("leina", "eater", "food");"#)
+        .expect_err("Expected validation error");
+    assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    let e = p
+        .load_str(r#"allow("leina", "food");"#)
+        .expect_err("Expected validation error");
+    assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    let e = p
+        .load_str(r#"allow_field("leina", "food");"#)
+        .expect_err("Expected validation error");
+    assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    let e = p
+        .load_str(r#"allow_request("leina", "eat", "food");"#)
+        .expect_err("Expected validation error");
+    assert!(matches!(e.kind, ErrorKind::Validation(_)));
+
+    // This should succeed
+    // TODO: should we emit warnings if rules with union specializers are loaded
+    // but no union types have been declared?
+    p.load_str(
+        r#"
+    has_permission(_actor: Actor, "eat", _resource: Resource);
+    has_permission(_actor: Actor, "eat", _resource: Actor);
+    has_role(_actor: Actor, "member", _resource: Resource);
+    has_role(_actor: Actor, "member", _resource: Actor);
+    has_relation(_actor: Actor, "any", _other: Actor);
+    has_relation(_actor: Resource, "any", _other: Actor);
+    has_relation(_actor: Resource, "any", _other: Resource);
+    has_relation(_actor: Actor, "any", _other: Resource);
+    allow("a", "b", "c");
+    allow_field("a", "b", "c", "d");
+    allow_request("a", "b");
+    "#,
+    )?;
+    // Make sure there are no warnings
+    assert!(p.next_message().is_none());
+    Ok(())
+}
